@@ -13,7 +13,8 @@ import docker
 
 from .helpers import which
 from .server import DATA_FOLDER
-from .rpc import BitcoinCLI, RpcError
+from .rpc import RpcError
+from .rpc_cache import BitcoinCLICached
 from .helpers import load_jsons
 
 
@@ -40,7 +41,7 @@ class Btcd_conn:
     def get_cli(self):
         ''' returns a BitcoinCLI '''
         # def __init__(self, user, passwd, host="127.0.0.1", port=8332, protocol="http", path="", timeout=30, **kwargs):
-        cli = BitcoinCLI(self.rpcuser, self.rpcpassword, host=self.ipaddress, port=self.rpcport)
+        cli = BitcoinCLICached(self.rpcuser, self.rpcpassword, host=self.ipaddress, port=self.rpcport)
         cli.getblockchaininfo()
         return cli
 
@@ -90,17 +91,17 @@ class BitcoindController:
     def testcoin_faucet(self, address, amount=20, mine_tx=False):
         ''' an easy way to get some testcoins '''
         cli = self.get_cli()
-        test3rdparty_cli = cli.wallet("test3rdparty")
         try:
-            balance = test3rdparty_cli.getbalance()
+            test3rdparty_cli = cli.wallet("test3rdparty")
         except RpcError as rpce:
             # return-codes:
             # https://github.com/bitcoin/bitcoin/blob/v0.15.0.1/src/rpc/protocol.h#L32L87
             if rpce.error_code == -18: # RPC_WALLET_NOT_FOUND
                 cli.createwallet("test3rdparty")
-                balance = test3rdparty_cli.getbalance()
+                test3rdparty_cli = cli.wallet("test3rdparty")
             else:
                 raise rpce
+        balance = test3rdparty_cli.getbalance()
         if balance < amount:
             test3rdparty_address = test3rdparty_cli.getnewaddress("test3rdparty")
             cli.generatetoaddress(102, test3rdparty_address)
@@ -252,7 +253,7 @@ class BitcoindDockerController(BitcoindController):
     def search_bitcoind_container(all=False):
         ''' returns a list of containers which are running bitcoind '''
         d_client = docker.from_env()
-        return [c for c in d_client.containers.list(all) if c.attrs['Config']['Cmd'][0]=='bitcoind']
+        return [c for c in d_client.containers.list(all) if (c.attrs['Config'].get('Cmd')or[""])[0]=='bitcoind']
 
     @staticmethod
     def detect_bitcoind_container(with_rpcport):
