@@ -1,4 +1,4 @@
-import sys, json, os, time, base64
+import ast, sys, json, os, time, base64
 import requests
 import random, copy
 from collections import OrderedDict
@@ -11,14 +11,13 @@ from flask import g, request, redirect, url_for
 from flask import Flask, Blueprint, render_template, request, redirect, url_for, jsonify, flash
 from flask_login import login_required, login_user, logout_user, current_user
 from flask_login.config import EXEMPT_METHODS
-from flask_qrcode import QRcode
 
 
 from .helpers import normalize_xpubs, run_shell
 from .descriptor import AddChecksum
-from .rpc import BitcoinCLI, RPC_PORTS
 
 from .logic import Specter, purposes, addrtypes, get_cli
+from .rpc import RpcError
 from datetime import datetime
 import urllib
 
@@ -33,19 +32,36 @@ from flask import current_app as app
 rand = random.randint(0, 1e32) # to force style refresh
 
 
+########## template injections #############
+@app.context_processor
+def inject_debug():
+    ''' Can be used in all jinja2 templates '''
+    print("DEBUG={}".format(app.config['DEBUG']))
+    return dict(debug=app.config['DEBUG'])
+
 ################ routes ####################
 
-@app.route('/combine/', methods=['GET', 'POST'])
+@app.route('/wallets/<wallet_alias>/combine/', methods=['GET', 'POST'])
 @login_required
-def combine():
+def combine(wallet_alias):
+    wallet = app.specter.wallets.get_by_alias(wallet_alias)
     if request.method == 'POST': # FIXME: ugly...
         d = request.json
         psbt0 = d['psbt0'] # request.args.get('psbt0')
         psbt1 = d['psbt1'] # request.args.get('psbt1')
-        psbt = app.specter.combine([psbt0, psbt1])
-        raw = app.specter.finalize(psbt)
+        txid = d['txid']
+        try:
+            psbt = app.specter.combine([psbt0, psbt1])
+            raw = app.specter.finalize(psbt)
+        except RpcError as e:
+            return e.error_msg, e.status_code
+        except Exception as e:
+            return "Unknown error: %r" % e, 500
         if "hex" in raw:
             app.specter.broadcast(raw["hex"])
+
+        device_name = d['device_name']
+        wallet.update_pending_psbt(psbt, txid, device_name)
         return json.dumps(raw)
     return 'meh'
 
@@ -227,6 +243,26 @@ def new_wallet_simple():
                 return render_template("base.html", error="Key not found", specter=app.specter, rand=rand)
             # create a wallet here
             wallet = app.specter.wallets.create_simple(wallet_name, wallet_type, key, device)
+            rescan_blockchain = 'rescanblockchain' in request.form
+            if rescan_blockchain:
+                if app.specter.info['chain'] == "main":
+                    if not app.specter.info['pruned'] or app.specter.info['pruneheight'] < 481824:
+                        startblock = 481824
+                    else:
+                        startblock = app.specter.info['pruneheight']
+                else:
+                    if not app.specter.info['pruned']:
+                        startblock = 0
+                    else:
+                        startblock = app.specter.info['pruneheight']
+                try:
+                    wallet.cli.rescanblockchain(startblock, timeout=1)
+                except requests.exceptions.ReadTimeout:
+                    pass
+                except Exception as e:
+                    print(e)
+                    error = "%r" % e
+                wallet.getdata()
             return redirect("/wallets/%s/" % wallet["alias"])
     return render_template("new_simple.html", wallet_name=wallet_name, device=device, error=err, specter=app.specter, rand=rand)
 
@@ -338,9 +374,8 @@ def wallet(wallet_alias):
 @login_required
 def wallet_tx(wallet_alias):
     app.specter.check()
-    try:
-        wallet = app.specter.wallets.get_by_alias(wallet_alias)
-    except:
+    wallet = app.specter.wallets.get_by_alias(wallet_alias)
+    if wallet is None:
         return render_template("base.html", error="Wallet not found", specter=app.specter, rand=rand)
 
     return render_template("wallet_tx.html", wallet_alias=wallet_alias, wallet=wallet, specter=app.specter, rand=rand)
@@ -353,13 +388,20 @@ def wallet_addresses(wallet_alias):
         wallet = app.specter.wallets.get_by_alias(wallet_alias)
     except:
         return render_template("base.html", error="Wallet not found", specter=app.specter, rand=rand)
+    viewtype = 'address' if request.args.get('view') != 'label' else 'label'
     if request.method == "POST":
         action = request.form['action']
         if action == "updatelabel":
             label = request.form['label']
-            address = request.form['addr']
-            wallet.setlabel(address, label)
-    return render_template("wallet_addresses.html", wallet_alias=wallet_alias, wallet=wallet, specter=app.specter, rand=rand)
+            account = request.form['account']
+            if viewtype == 'address':
+                wallet.setlabel(account, label)
+            else:
+                for address in wallet.addressesonlabel(account):
+                    wallet.setlabel(address, label)
+                wallet.getdata()
+    alladdresses = True if request.args.get('all') != 'False' else False
+    return render_template("wallet_addresses.html", wallet_alias=wallet_alias, wallet=wallet, alladdresses=alladdresses, viewtype=viewtype, specter=app.specter, rand=rand)
 
 @app.route('/wallets/<wallet_alias>/receive/', methods=['GET', 'POST'])
 @login_required
@@ -386,7 +428,7 @@ def fees(blocks):
     res = app.specter.estimatesmartfee(int(blocks))
     return res
 
-@app.route('/wallets/<wallet_alias>/send/', methods=['GET', 'POST'])
+@app.route('/wallets/<wallet_alias>/send/new', methods=['GET', 'POST'])
 @login_required
 def wallet_send(wallet_alias):
     app.specter.check()
@@ -408,10 +450,11 @@ def wallet_send(wallet_alias):
             label = request.form['label']
             if request.form['label'] != "":
                 wallet.setlabel(address, label)
-            amount = float(request.form['amount'])
+            amount = float(request.form['btc_amount'])
             subtract = bool(request.form.get("subtract", False))
             fee_unit = request.form.get('fee_unit')
-
+            selected_coins = request.form.getlist('coinselect')
+            print("selected coins: {}".format(selected_coins))
             if 'dynamic' in request.form.get('fee_options'):
                 fee_rate = float(request.form.get('fee_rate_dynamic'))
             else:
@@ -419,7 +462,7 @@ def wallet_send(wallet_alias):
                     fee_rate = float(request.form.get('fee_rate'))
 
             try:
-                psbt = wallet.createpsbt(address, amount, subtract=subtract, fee_rate=fee_rate, fee_unit=fee_unit)
+                psbt = wallet.createpsbt(address, amount, subtract=subtract, fee_rate=fee_rate, fee_unit=fee_unit, selected_coins=selected_coins)
                 if psbt is None:
                     err = "Probably you don't have enough funds, or something else..."
                 else:
@@ -429,11 +472,47 @@ def wallet_send(wallet_alias):
                             if address in v["scriptPubKey"]["addresses"]:
                                 amount = v["value"]
             except Exception as e:
-                err = "%r" % e
-    return render_template("wallet_send.html", psbt=psbt, address=address, label=label, amount=amount, 
+                err = e
+            if err is None:
+                return render_template("wallet_send_sign_psbt.html", psbt=psbt, label=label, 
+                                                    wallet_alias=wallet_alias, wallet=wallet, 
+                                                    specter=app.specter, rand=rand)
+        elif action == "openpsbt":
+            psbt = ast.literal_eval(request.form["pending_psbt"])
+            return render_template("wallet_send_sign_psbt.html", psbt=psbt, label=label, 
+                                                wallet_alias=wallet_alias, wallet=wallet, 
+                                                specter=app.specter, rand=rand)
+        elif action == 'deletepsbt':
+            try:
+                wallet.delete_pending_psbt(ast.literal_eval(request.form["pending_psbt"])["tx"]["txid"])
+            except Exception as e:
+                flash("Could not delete Pending PSBT!")
+    return render_template("wallet_send.html", psbt=psbt, label=label, 
                                                 wallet_alias=wallet_alias, wallet=wallet, 
                                                 specter=app.specter, rand=rand, error=err)
 
+@app.route('/wallets/<wallet_alias>/send/pending/', methods=['GET', 'POST'])
+@login_required
+def wallet_sendpending(wallet_alias):
+    app.specter.check()
+    try:
+        wallet = app.specter.wallets.get_by_alias(wallet_alias)
+    except Exception as e:
+        print(e)
+        return render_template("base.html", error="Wallet not found", specter=app.specter, rand=rand)
+    if request.method == "POST":
+        action = request.form['action']
+        if action == 'deletepsbt':
+            try:
+                wallet.delete_pending_psbt(ast.literal_eval(request.form["pending_psbt"])["tx"]["txid"])
+            except Exception as e:
+                flash("Could not delete Pending PSBT!")
+    pending_psbts = wallet.pending_psbts
+    return render_template("wallet_sendpending.html", pending_psbts=pending_psbts,
+                                                wallet_alias=wallet_alias, wallet=wallet, 
+                                                specter=app.specter) 
+
+
 @app.route('/wallets/<wallet_alias>/settings/', methods=['GET','POST'])
 @login_required
 def wallet_settings(wallet_alias):
@@ -465,6 +544,18 @@ def wallet_settings(wallet_alias):
             wallet.keypoolrefill(wallet["keypool"], wallet["keypool"]+delta)
             wallet.keypoolrefill(wallet["change_keypool"], wallet["change_keypool"]+delta, change=True)
             wallet.getdata()
+        elif action == "rebuildcache":
+            wallet.cli.cache.rebuild_cache()
+        elif action == "deletewallet":
+            app.specter.wallets.delete_wallet(wallet)
+            response = redirect(url_for('index'))
+            return response
+        elif action == "renamewallet":
+            wallet_name = request.form['walletname']
+            if wallet_name in app.specter.wallets.names():
+                error = "Wallet already exists"
+            else:
+                app.specter.wallets.rename_wallet(wallet, wallet_name)
 
     cc_file = None
     qr_text = wallet["name"]+"&"+wallet.descriptor
@@ -589,5 +680,9 @@ def derivation(wallet):
 def txonaddr(obj):
     return json.dumps(obj, indent=4)
 
+@app.template_filter('btcamount')
+def btcamount(value):
+    value = float(value)
+    return "{:.8f}".format(value).rstrip("0").rstrip(".")
 
-    
\ No newline at end of file
+    
