@@ -2,16 +2,20 @@ import base64
 import copy
 import json
 import os
+import shutil
 import random
 import hashlib
+from time import time
 from collections import OrderedDict
 
 from . import helpers
 from .descriptor import AddChecksum
-from .helpers import deep_update, load_jsons
-from .rpc import RPC_PORTS, BitcoinCLI, autodetect_cli
+from .helpers import deep_update, load_jsons, get_xpub_fingerprint
+from .rpc import RPC_PORTS, autodetect_cli_confs, get_default_datadir
+from .rpc_cache import BitcoinCLICached
 from .serializations import PSBT
 
+
 # a gap of 20 addresses is what many wallets do
 WALLET_CHUNK = 20
 # we don't need to scan earlier than that 
@@ -46,15 +50,15 @@ def get_cli(conf):
         conf["autodetect"] = True
     if conf["autodetect"]:
         if "port" in conf:
-            cli_arr = autodetect_cli(port=conf["port"])
+            cli_conf_arr = autodetect_cli_confs(port=conf["port"])
         else:
-            cli_arr = autodetect_cli()
-        if len(cli_arr) > 0:
-            cli = cli_arr[0]
+            cli_conf_arr = autodetect_cli_confs()
+        if len(cli_conf_arr) > 0:
+            cli = BitcoinCLICached(**cli_conf_arr[0])
         else:
             return None
     else:
-        cli = BitcoinCLI(conf["user"], conf["password"], 
+        cli = BitcoinCLICached(conf["user"], conf["password"], 
                           host=conf["host"], port=conf["port"], protocol=conf["protocol"])
     return cli
 
@@ -127,7 +131,7 @@ class Specter:
         self._is_running = False
         if self._is_configured:
             try:
-                self._info = self.cli.getmininginfo()
+                self._info = self.cli.getblockchaininfo()
                 self._is_running = True
             except:
                 pass
@@ -137,7 +141,7 @@ class Specter:
 
         chain = self._info["chain"]
         if self.wallets is None or chain is None:
-            wallets_path = "specter%s/" % self.config["uid"]
+            wallets_path = "specter%s" % self.config["uid"]
             self.wallets = WalletManager(
                                 os.path.join(self.data_folder, "wallets"), 
                                 self.cli, 
@@ -301,7 +305,7 @@ class DeviceManager:
         return self
 
     def __next__(self):
-        arr = list(self._devices.keys())
+        arr = sorted(list(self._devices.keys()))
         if self._n < len(arr):
             v = self._devices[arr[self._n]]
             self._n += 1
@@ -346,7 +350,7 @@ class Device(dict):
 
 class WalletManager:
     # chain is required to manage wallets when bitcoin-cli is not running
-    def __init__(self, data_folder, cli, chain, path="specter/"):
+    def __init__(self, data_folder, cli, chain, path="specter"):
         self.data_folder = data_folder
         self.chain = chain
         self.cli = cli
@@ -375,7 +379,7 @@ class WalletManager:
             wallets = load_jsons(self.working_folder, key="name")
             existing_wallets = [w["name"] for w in self.cli.listwalletdir()["wallets"]]
             for k in wallets:
-                if self.cli_path+wallets[k]["alias"] in existing_wallets:
+                if os.path.join(self.cli_path,wallets[k]["alias"]) in existing_wallets:
                     self._wallets[k] = wallets[k]
         else:
             self._wallets = {}
@@ -386,22 +390,27 @@ class WalletManager:
         not_loaded_wallets = [w for w in loadable_wallets if w not in loaded_wallets]
         # print("not loaded wallets:", not_loaded_wallets)
         for k in self._wallets:
-            if self.cli_path+self._wallets[k]["alias"] in not_loaded_wallets:
+            if os.path.join(self.cli_path,self._wallets[k]["alias"]) in not_loaded_wallets:
                 print("loading", self._wallets[k]["alias"])
-                self.cli.loadwallet(self.cli_path+self._wallets[k]["alias"])
+                self.cli.loadwallet(os.path.join(self.cli_path,self._wallets[k]["alias"]))
+                if "pending_psbts" in self._wallets[k] and len(self._wallets[k]["pending_psbts"]) > 0:
+                    for psbt in self._wallets[k]["pending_psbts"]:
+                        print("lock", self._wallets[k]["alias"], self._wallets[k]["pending_psbts"][psbt]["tx"]["vin"])
+                        Wallet(self._wallets[k], self).cli.lockunspent(False, [utxo for utxo in self._wallets[k]["pending_psbts"][psbt]["tx"]["vin"]])
 
-    def get_by_alias(self, fname):
-        for dev in self:
-            if dev["alias"] == fname:
-                return dev
+    def get_by_alias(self, alias):
+        for w in self:
+            if w["alias"] == alias:
+                return w
 
     def names(self):
         return list(self._wallets.keys())
 
     def _get_intial_wallet_dict(self, name):
+        walletsindir = [wallet["name"] for wallet in self.cli.listwalletdir()["wallets"]]
         al = alias(name)
         i = 2
-        while os.path.isfile(os.path.join(self.working_folder, "%s.json" % al)):
+        while os.path.isfile(os.path.join(self.working_folder, "%s.json" % al)) or os.path.join(self.cli_path,al) in walletsindir:
             al = alias("%s %d" % (name, i))
             i+=1
         dic = {
@@ -414,6 +423,7 @@ class WalletManager:
             "change_index": 0,
             "change_address": None,
             "change_keypool": 0,
+            "pending_psbts": {}
         }
         return dic
 
@@ -443,7 +453,7 @@ class WalletManager:
         # add wallet to internal dict
         self._wallets[o["alias"]] = o
         # create a wallet in Bitcoin Core
-        r = self.cli.createwallet(self.cli_path+o["alias"], True)
+        r = self.cli.createwallet(os.path.join(self.cli_path,o["alias"]), True)
         # save wallet file to disk
         if self.working_folder is not None:
             with open(o["fullpath"], "w+") as f:
@@ -491,7 +501,7 @@ class WalletManager:
         # add wallet to internal dict
         self._wallets[o["alias"]] = o
         # create a wallet in Bitcoin Core
-        r = self.cli.createwallet(self.cli_path+o["alias"], True)
+        r = self.cli.createwallet(os.path.join(self.cli_path,o["alias"]), True)
         # save wallet file to disk
         if self.working_folder is not None:
             with open(o["fullpath"], "w+") as f:
@@ -502,6 +512,24 @@ class WalletManager:
         w = Wallet(o, self)
         return w
 
+    def delete_wallet(self, wallet):
+        print("Deleting {}".format(wallet["alias"]))
+        self.cli.unloadwallet(os.path.join(self.cli_path,wallet["alias"]))
+        # Try deleting wallet file
+        if get_default_datadir() and os.path.exists(os.path.join(get_default_datadir(), os.path.join(self.cli_path,wallet["alias"]))):
+            shutil.rmtree(os.path.join(get_default_datadir(), os.path.join(self.cli_path,wallet["alias"])))
+        # Delete JSON
+        if os.path.exists(wallet["fullpath"]):
+            os.remove(wallet["fullpath"])
+
+    def rename_wallet(self, wallet, name):
+        print("Renaming {}".format(wallet["alias"]))
+        wallet["name"] = name
+        if self.working_folder is not None:
+            with open(wallet["fullpath"], "w+") as f:
+                f.write(json.dumps(wallet, indent=4))
+        self.update()
+
     def __getitem__(self, name):
         return Wallet(self._wallets[name], manager=self)
 
@@ -510,7 +538,7 @@ class WalletManager:
         return self
 
     def __next__(self):
-        arr = list(self._wallets.keys())
+        arr = sorted(list(self._wallets.keys()))
         if self._n < len(arr):
             v = self._wallets[arr[self._n]]
             self._n += 1
@@ -526,7 +554,7 @@ class Wallet(dict):
         self.update(d)
         self.manager = manager
         self.cli_path = manager.cli_path
-        self.cli = manager.cli.wallet(self.cli_path+self["alias"])
+        self.cli = manager.cli.wallet(os.path.join(self.cli_path,self["alias"]))
         self._dict = d
         # check if address is known and derive if not
         # address derivation will also refill the keypool if necessary
@@ -535,6 +563,7 @@ class Wallet(dict):
             self.setlabel(self._dict["address"], "Address #0")
         if self._dict["change_address"] is None:
             self._dict["change_address"] = self.get_address(0, change=True)
+        self.cli.scan_addresses(self)
         self.getdata()
 
     def _commit(self, update_manager=True):
@@ -567,6 +596,45 @@ class Wallet(dict):
     def is_multisig(self):
         return "sigs_required" in self
 
+    @property
+    def sigs_required(self):
+        return 1 if not self.is_multisig else self["sigs_required"]
+
+    @property
+    def pending_psbts(self):
+        if "pending_psbts" not in self._dict:
+            return {}
+        return self._dict["pending_psbts"]
+
+    @property
+    def locked_amount(self):
+        amount = 0
+        for psbt in self.pending_psbts:
+            amount += sum([utxo["witness_utxo"]["amount"] for utxo in self.pending_psbts[psbt]["inputs"]])
+        return amount
+
+    def delete_pending_psbt(self, txid):
+        try:
+            self.cli.lockunspent(True, self._dict["pending_psbts"][txid]["tx"]["vin"])
+        except:
+            # UTXO was spent
+            pass
+        del self._dict["pending_psbts"][txid]
+        self._commit()
+
+    def update_pending_psbt(self, psbt, txid, device_name):
+        if txid in self._dict["pending_psbts"]:
+            if self._dict["pending_psbts"][txid]["sigs_count"] + 1 == self.sigs_required:
+                self.delete_pending_psbt(txid)
+                return
+            self._dict["pending_psbts"][txid]["sigs_count"] += 1
+            self._dict["pending_psbts"][txid]["base64"] = psbt
+            if device_name:
+                if "devices_signed" not in self._dict["pending_psbts"][txid]:
+                    self._dict["pending_psbts"][txid]["devices_signed"] = []
+                self._dict["pending_psbts"][txid]["devices_signed"].append(device_name)
+            self._commit()
+
     def _check_change(self):
         addr = self["change_address"]
         if addr is not None:
@@ -583,23 +651,42 @@ class Wallet(dict):
         self._dict["change_address"] = self.get_address(self._dict["change_index"], change=True)
         self._commit()
 
+    @property
+    def txlist(self):
+        ''' The last 1000 transactions for that wallet - filtering out change addresses transactions and duplicated transactions (except for self-transfers)
+            This list is used for the wallet `txs` tab to list the wallet transacions.
+        '''
+        txidlist = []
+        txlist = []
+        for tx in self.transactions:
+            if tx["is_change"] == False and (tx["is_self"] or tx["txid"] not in txidlist):
+                txidlist.append(tx["txid"])
+                txlist.append(tx)
+        return txlist
+
     def getdata(self):
         try:
             self.balance = self.getbalances()
         except:
             self.balance = None
         try:
-            self.transactions = self.cli.listtransactions("*", 1000, 0, True)[::-1]
+            self.utxo = self.cli.listunspent(0)
+        except:
+            self.utxo = None
+        try:
+            self.transactions = self.cli.listtransactions("*", 1000, 0, True)
         except:
             self.transactions = None
         try:
             self.info = self.cli.getwalletinfo()
         except:
             self.info = None
+
         self._check_change()
         return {
             "balance": self.balance,
-            "transactions": self.transactions
+            "transactions": self.transactions,
+            "utxo": self.utxo
         }
 
     @property
@@ -612,11 +699,14 @@ class Wallet(dict):
         else:
             return self.info["scanning"]["progress"]
 
-    def getnewaddress(self):
-        self._dict["address_index"] += 1
-        addr = self.get_address(self._dict["address_index"])
-        self.setlabel(addr, "Address #{}".format(self._dict["address_index"]))
-        self._dict["address"] = addr
+    def getnewaddress(self, change=False):
+        label = "Change" if change else "Address"
+        index_type = "change_index" if change else "address_index"
+        address_type = "change_address" if change else "address"
+        self._dict[index_type] += 1
+        addr = self.get_address(self._dict[index_type], change=change)
+        self.setlabel(addr, "{} #{}".format(label, self._dict[index_type]))
+        self._dict[address_type] = addr
         self._commit()
         return addr
 
@@ -629,9 +719,9 @@ class Wallet(dict):
             self.keypoolrefill(self._dict[pool], index+WALLET_CHUNK, change=change)
         if self.is_multisig:
             # using sortedmulti for addresses
-            sorted_desc = self.sort_descriptor(self[desc], index)
-            return self.cli.deriveaddresses(sorted_desc)[0]
-        return self.cli.deriveaddresses(self._dict[desc], [index, index+1])[0]
+            sorted_desc = self.sort_descriptor(self[desc], index=index, change=change)
+            return self.cli.deriveaddresses(sorted_desc, change=change)[0]
+        return self.cli.deriveaddresses(self._dict[desc], [index, index+1], change=change)[0]
 
     def geterror(self):
         if self.cli.r is not None:
@@ -668,6 +758,15 @@ class Wallet(dict):
             unspent = self.cli.listunspent(0, 0)
             for t in unspent:
                 r["untrusted_pending"] += t["amount"]
+            # Bitcoin Core doesn't return locked UTXO with `listunspent` command. 
+            # `getbalance` command doesn't return balance from unconfirmed UTXO.
+            # So to imitate `getbalances` we need to add all unconfirmed locked UTXO balance manually with this loop.
+            locked_utxo = self.cli.listlockunspent()
+            for tx in locked_utxo:
+                tx_data = self.cli.gettransaction(tx["txid"])
+                raw_tx = self.cli.decoderawtransaction(tx_data["hex"])
+                if "confirmations" not in tx_data or tx_data["confirmations"] == 0:
+                    r["untrusted_pending"] += raw_tx["vout"][tx["vout"]]["value"]
         except:
             r = { "trusted": None, "untrusted_pending": None }
         self.balance = r
@@ -680,14 +779,14 @@ class Wallet(dict):
             return None
         return r["trusted"]+r["untrusted_pending"]
 
-    def sort_descriptor(self, descriptor, index=None):
+    def sort_descriptor(self, descriptor, index=None, change=False):
         if index is not None:
             descriptor = descriptor.replace("*", f"{index}")
         # remove checksum
         descriptor = descriptor.split("#")[0]
 
         # get address (should be already imported to the wallet)
-        address = self.cli.deriveaddresses(AddChecksum(descriptor))[0]
+        address = self.cli.deriveaddresses(AddChecksum(descriptor), change=change)[0]
 
         # get pubkeys involved
         address_info = self.cli.getaddressinfo(address)
@@ -740,7 +839,7 @@ class Wallet(dict):
             # we do one at a time
             args[0].pop("range")
             for i in range(start, end):
-                sorted_desc = self.sort_descriptor(self[desc], i)
+                sorted_desc = self.sort_descriptor(self[desc], index=i, change=change)
                 args[0]["desc"] = sorted_desc
                 self.cli.importmulti(args, {"rescan": False}, timeout=120)
         self._dict[pool] = end
@@ -751,8 +850,32 @@ class Wallet(dict):
         txlist = [tx for tx in self.transactions if tx["address"] == addr]
         return len(txlist)
 
+    def balanceonaddr(self, addr):
+        balancelist = [utxo["amount"] for utxo in self.utxo if utxo["address"] == addr]
+        return sum(balancelist)
+
+    def txonlabel(self, label):
+        txlist = [tx for tx in self.transactions if self.getlabel(tx["address"]) == label]
+        return len(txlist)
+
+    def balanceonlabel(self, label):
+        balancelist = [utxo["amount"] for utxo in self.utxo if self.getlabel(utxo["address"]) == label]
+        return sum(balancelist)
+
+    def addressesonlabel(self, label):
+        return list(dict.fromkeys(
+            [tx["address"] for tx in self.transactions if self.getlabel(tx["address"]) == label]
+        ))
+
+    def istxspent(self, txid):
+        return txid in [utxo["txid"] for utxo in self.utxo]
+
     def setlabel(self, addr, label):
         self.cli.setlabel(addr, label)
+
+    def getlabel(self, addr):
+        address_info = self.cli.getaddressinfo(addr)
+        return address_info["label"] if "label" in address_info and address_info["label"] != "" else addr
     
     def getaddressname(self, addr, addr_idx):
         address_info = self.cli.getaddressinfo(addr)
@@ -761,7 +884,7 @@ class Wallet(dict):
             address_info["label"] = "Address #{}".format(addr_idx)
         return addr if ("label" not in address_info or address_info["label"] == "") else address_info["label"]
 
-    @property    
+    @property
     def fullbalance(self):
         ''' This is cached. Consider to use getfullbalance '''
         if self.balance is None:
@@ -770,6 +893,21 @@ class Wallet(dict):
             return None
         return self.balance["trusted"]+self.balance["untrusted_pending"]
 
+    @property
+    def availablebalance(self):
+        ''' This is cached.'''
+        if self.balance is None:
+            return None
+        if self.balance["trusted"] is None or self.balance["untrusted_pending"] is None:
+            return None
+        locked = [0]
+        for psbt in self.pending_psbts:
+            for i, txid in enumerate([tx["txid"] for tx in self.pending_psbts[psbt]["tx"]["vin"]]):
+                tx_data = self.cli.gettransaction(txid)
+                if "confirmations" in tx_data and tx_data["confirmations"] != 0:
+                    locked.append(self.pending_psbts[psbt]["inputs"][i]["witness_utxo"]["amount"])
+        return self.balance["trusted"]+self.balance["untrusted_pending"] - sum(locked)
+
     @property
     def descriptor(self):
         return self['recv_descriptor'].split("#")[0].replace("/0/*", "").replace("multi", "sortedmulti")
@@ -786,11 +924,47 @@ class Wallet(dict):
         addr = self["address"]
         return self.txonaddr(addr)
 
+    @property
+    def utxoaddresses(self):
+        return list(dict.fromkeys([
+            utxo["address"] for utxo in 
+            sorted(
+                self.utxo,
+                key = lambda utxo: next(
+                    tx for tx in self.transactions if tx["txid"] == utxo["txid"]
+                )["time"]
+            )
+        ]))
+
+    @property
+    def utxolabels(self):
+        return list(dict.fromkeys([
+            self.getlabel(utxo["address"]) for utxo in 
+            sorted(
+                self.utxo,
+                key = lambda utxo: next(
+                    tx for tx in self.transactions if tx["txid"] == utxo["txid"]
+                )["time"]
+            )
+        ]))
+
     @property
     def addresses(self):
         return [self.get_address(idx) for idx in range(0,self._dict["address_index"] + 1)]
 
-    def createpsbt(self, address:str, amount:float, subtract:bool=False, fee_rate:float=0.0, fee_unit="SAT_B"):
+    @property
+    def active_addresses(self):
+        return list(dict.fromkeys(self.addresses + self.utxoaddresses))
+    
+    @property
+    def change_addresses(self):
+        return [self.get_address(idx, change=True) for idx in range(0,self._dict["change_index"] + 1)]
+
+    @property
+    def labels(self):
+        return list(dict.fromkeys([self.getlabel(addr) for addr in self.active_addresses]))
+
+    def createpsbt(self, address:str, amount:float, subtract:bool=False, fee_rate:float=0.0, fee_unit="SAT_B", selected_coins=[]):
         """
             fee_rate: in sat/B or BTC/kB. Default (None) bitcoin core sets feeRate automatically.
         """
@@ -809,6 +983,18 @@ class Wallet(dict):
                 b -= tx["amount"]
                 if b < 0:
                     break;
+        elif selected_coins != []:
+            still_needed = amount
+            for coin in selected_coins:
+                coin_txid = coin.split(",")[0]
+                coin_vout = int(coin.split(",")[1])
+                coin_amount = float(coin.split(",")[2])
+                extra_inputs.append({"txid": coin_txid, "vout": coin_vout})
+                still_needed -= coin_amount
+                if still_needed < 0:
+                    break;
+            if still_needed > 0:
+                raise SpecterError("Selected coins does not cover Full amount! Please select more coins!")
 
         # subtract fee from amount of this output:
         # currently only one address is supported, so either
@@ -821,6 +1007,8 @@ class Wallet(dict):
             "subtractFeeFromOutputs": subtract_arr
         }
 
+        self.setlabel(self["change_address"], "Change #{}".format(self._dict["change_index"]))
+
         if fee_rate > 0.0 and fee_unit == "SAT_B":
             # bitcoin core needs us to convert sat/B to BTC/kB
             options["feeRate"] = fee_rate / 10 ** 8 * 1024
@@ -842,7 +1030,15 @@ class Wallet(dict):
         if self.is_multisig:
             for k in self._dict["keys"]:
                 key = b'\x01'+helpers.decode_base58(k["xpub"])
-                value = bytes.fromhex(k["fingerprint"])+der_to_bytes(k["derivation"])
+                if "fingerprint" in k and k["fingerprint"] is not None:
+                    fingerprint = bytes.fromhex(k["fingerprint"])
+                else:
+                    fingerprint = helpers.get_xpub_fingerprint(k["xpub"])
+                if "derivation" in k and k["derivation"] is not None:
+                    der = der_to_bytes(k["derivation"])
+                else:
+                    der = b''
+                value = fingerprint+der
                 cc_psbt.unknown[key] = value
         psbt["coldcard"]=cc_psbt.serialize()
 
@@ -861,6 +1057,18 @@ class Wallet(dict):
                 inp.hd_keypaths = {}
         psbt["specter"]=qr_psbt.serialize()
         print("PSBT for Specter:", psbt["specter"])
+
+        self.cli.lockunspent(False, psbt["tx"]["vin"])
+
+        if "pending_psbts" not in self._dict:
+            self._dict["pending_psbts"] = {}
+        psbt["amount"] = amount
+        psbt["address"] = address
+        psbt["time"] = time()
+        psbt["sigs_count"] = 0
+        self._dict["pending_psbts"][psbt["tx"]["txid"]] = psbt
+        self._commit()
+
         return psbt
 
     def get_cc_file(self):
@@ -873,7 +1081,7 @@ class Wallet(dict):
         # cc assume the same derivation for all keys :(
         derivation = None
         for k in self["keys"]:
-            if "derivation" in k:
+            if "derivation" in k and k["derivation"] is not None:
                 derivation = k["derivation"].replace("h","'")
                 break
         if derivation is None:
@@ -894,7 +1102,7 @@ Format: {}
             if 'fingerprint' in k:
                 fingerprint = k['fingerprint']
             if fingerprint is None:
-                return None
+                fingerprint = get_xpub_fingerprint(k['xpub']).hex()
             cc_file += "{}: {}\n".format(fingerprint.upper(), k['xpub'])
         return cc_file
 
