@@ -16,17 +16,17 @@ rand = random.randint(0, 1e32) # to force style refresh
 
 hwi_views = Blueprint('hwi', __name__, template_folder='templates')
 
-def get_spector_instance():
+def get_specter_instance():
     # specter instance is injected into app in server.py's __main__()
     return current_app.specter
 
 
-def get_hwi_client(type, path):
-    is_test = 'test' in get_spector_instance().chain
+def get_hwi_client(type, path, passphrase=''):
+    is_test = 'test' in get_specter_instance().chain
     if type == "specter":
         client = SpecterClient(path)
     else:
-        client = hwilib_commands.get_client(type, path)
+        client = hwilib_commands.get_client(type, path, passphrase)
     client.is_testnet = is_test
     return client
 
@@ -56,7 +56,7 @@ def _enumerate():
 
 @hwi_views.route('/extract_xpubs/', methods=['POST'])
 def hwi_extract_xpubs():
-    specter = get_spector_instance()
+    specter = get_specter_instance()
 
     device_name = request.form['device_name']
     if device_name in specter.devices.names():
@@ -64,9 +64,10 @@ def hwi_extract_xpubs():
     
     type = request.form.get("type")
     path = request.form.get("path")
+    passphrase = request.form.get("passphrase")
 
     try:
-        client = get_hwi_client(type, path)
+        client = get_hwi_client(type, path, passphrase=passphrase)
 
         # Client will be configured for testnet if our Specter instance is
         #   currently connected to testnet. This will prevent us from
@@ -74,8 +75,7 @@ def hwi_extract_xpubs():
         client.is_testnet = False
         xpubs = ""
 
-        master_xpub = client.get_pubkey_at_path('m/0')['xpub']
-        master_fpr = hwilib_commands.get_xpub_fingerprint_hex(master_xpub)
+        master_fpr = client.get_master_fingerprint_hex()
 
         # HWI calls to client.get_pubkey_at_path() return "xpub"-prefixed xpubs
         # regardless of derivation path. Update to match SLIP-0132 prefixes.
@@ -151,7 +151,7 @@ def hwi_extract_xpubs():
 @hwi_views.route('/new_device/', methods=['GET'])
 def hwi_new_device_xpubs():
     err = None
-    specter = get_spector_instance()
+    specter = get_specter_instance()
     specter.check()
 
     return render_template(
@@ -201,6 +201,8 @@ def hwi_prompt_pin():
     print(request.form)
     type = request.form.get("type")
     path = request.form.get("path")
+    passphrase = request.form.get("passphrase")
+    print('passphrase: {}'.format(passphrase))
 
     try:
         if type == "keepkey" or type == "trezor":
@@ -210,7 +212,7 @@ def hwi_prompt_pin():
             #       7 8 9
             #       4 5 6
             #       1 2 3
-            client = get_hwi_client(type, path)
+            client = get_hwi_client(type, path, passphrase=passphrase)
             status = hwilib_commands.prompt_pin(client)
             return jsonify(success=True, status=status)
         else:
@@ -224,10 +226,11 @@ def hwi_prompt_pin():
 def hwi_send_pin():
     type = request.form.get("type")
     path = request.form.get("path")
+    passphrase = request.form.get("passphrase")
     pin = request.form.get("pin")
 
     try:
-        client = get_hwi_client(type, path)
+        client = get_hwi_client(type, path, passphrase=passphrase)
         status = hwilib_commands.send_pin(client, pin)
         return jsonify(status)
     except Exception as e:
@@ -239,10 +242,11 @@ def hwi_send_pin():
 def hwi_sign_tx():
     type = request.form.get("type")
     path = request.form.get("path")
+    passphrase = request.form.get("passphrase")
     psbt = request.form.get("psbt")
 
     try:
-        client = get_hwi_client(type, path)
+        client = get_hwi_client(type, path, passphrase=passphrase)
         status = hwilib_commands.signtx(client, psbt)
         print(status)
 
@@ -261,3 +265,19 @@ def hwi_sign_tx():
 
         return jsonify(success=False, error=str(e))
 
+@hwi_views.route('/display_address/', methods=['POST'])
+def hwi_display_address():
+    type = request.form.get("type")
+    path = request.form.get("path")
+    passphrase = request.form.get("passphrase")
+    descriptor = request.form.get("descriptor")
+    
+    try:
+        client = get_hwi_client(type, path, passphrase=passphrase)
+        status = hwilib_commands.displayaddress(client, desc=descriptor)
+        if 'error' in status:
+            return jsonify(success=False, error=status['error'])
+        return jsonify(success=True, status=status)
+    except Exception as e:
+        print(e)
+        return jsonify(success=False, error=e)
